import json
from http import HTTPStatus

from aws_lambda_powertools.event_handler import Response, content_types
from aws_lambda_powertools.event_handler.exceptions import BadRequestError

from aws_packages.auth.auth_backend_icp import icp_auth_backend
from aws_packages.auth.auth_backend_solana import solana_auth_backend
from aws_packages.auth.models import AuthenticationRequest, LoginResponse, User


def process_authenticate_request(request_body, auth_backend):
    if request_body is None:
        raise BadRequestError("No request body provided")

    request_obj = AuthenticationRequest(**json.loads(request_body))

    try:
        auth_user = auth_backend.authenticate(request_obj)
    except Exception as e:
        return Response(
            status_code=HTTPStatus.BAD_REQUEST.value,
            content_type=content_types.APPLICATION_JSON,
            body={"message": str(e)},
        )

    return Response(
        status_code=HTTPStatus.OK.value,
        content_type=content_types.APPLICATION_JSON,
        body=LoginResponse(
            access_token=auth_backend.get_access_token(auth_user)
        ).as_dict(),
    )


def process_icp_authenticate_request(request_body):
    return process_authenticate_request(request_body, icp_auth_backend)


def process_solana_authenticate_request(request_body):
    return process_authenticate_request(request_body, solana_auth_backend)



import boto3
from boto3.dynamodb.conditions import Key

_dynamodb_client = boto3.resource("dynamodb")


def get_dynamodb_client():
    return _dynamodb_client


class DataHandler:
    def __init__(self, table_name, logger=None, client=None, key="pk"):
        self.key = key
        if client is None:
            self._client = get_dynamodb_client()
        else:
            self._client = client
        self._table_name = table_name
        self._logger = logger
        self.table = self._client.Table(self._table_name)

    def add_item(self, identifier, value):
        self.table.put_item(Item={self.key: identifier, **value})

    def update_item(self, identifier, value):
        if value.has_key(self.key):
            self.log("remove key from the value")
            value.pop(self.key)
        self.table.put_item(Item={self.key: identifier, **value})

    def log(self, message):
        if self._logger is None:
            print(message)
        self._logger.info(msg=message)

    def get_item(self, identifier):
        response = self.table.query(KeyConditionExpression=Key(self.key).eq(identifier))
        items = response["Items"]

        if not items:
            return None

        if len(items) > 1:
            self.log("Multiple items present in the queryset")
            raise Exception(f"More than one item found for {identifier}")

        entry = items[0]
        return entry

    def scan_items(self, *args, **kwargs):
        result = self.table.scan(**kwargs)
        return result["Items"]

    def query_items(self, *args, **kwargs):
        result = self.table.query(**kwargs)
        if not result:
            return None
        return result["Items"]



"""Authentication backends.

"""

import json
import os
from http import HTTPStatus

from aws_lambda_powertools.event_handler import Response, content_types
from aws_lambda_powertools.event_handler.exceptions import BadRequestError
from ic.agent import Agent
from ic.candid import Types, encode
from ic.client import Client
from ic.identity import Identity

from aws_packages.auth.auth_backend_base import AuthBackendBase
from aws_packages.auth.models import AuthenticationRequest, LoginResponse, User
from aws_packages.auth.tokens import AccessToken


class ICPAuthBackend(AuthBackendBase):
    """Authentication backend with ICP Canister"""

    def __init__(self, url: str, auth_canister: str, auth_function: str):
        self._url = url
        self._identity = Identity()
        self._client = Client(url=self._url)
        self._icp_agent = Agent(self._identity, self._client)
        self._auth_canister = auth_canister
        self._canister_auth_func = auth_function

    def authenticate(self, login_request: AuthenticationRequest) -> User:
        """Authenticate the request and returns an authenticated User"""
        encoded_args = encode(
            [
                {
                    "type": Types.Principal,
                    "value": login_request.user,
                }
            ]
        )

        self._icp_agent.update_raw(
            self._auth_canister, self._canister_auth_func, encoded_args
        )

        return User(principal=login_request.user)

    def authenticate_with_token(self, token: str):
        """Authenticate user using jwt token and return status"""
        print("token : ", token)
        AccessToken(token_string=token)

    def get_access_token(self, user: User) -> str:
        """Get the access for the given user"""
        token = AccessToken()
        return token.get_access_token(user)


icp_auth_backend = ICPAuthBackend(
    url="https://ic0.app",
    auth_canister=os.environ.get("CANISTER_ID"),
    auth_function="getUserToken",
)


if __name__ == "__main__":
    icp_auth_backend = ICPAuthBackend(
        url="https://ic0.app",
        auth_canister="6qy4q-5aaaa-aaaah-adwma-cai",
        auth_function="getUserToken",
    )

    request = AuthenticationRequest(
        user="4cay5-ew3bs-vr6yl-7iffu-67doc-l655v-dluy7-qplpx-7pkio-er5rt-uqe",
        authcode="",
    )
    user = icp_auth_backend.authenticate(request)
    print(user)



from dataclasses import dataclass


@dataclass
class AuthorizationRequest:
    """Authorization request model"""

    Authorization: str

    @property
    def token(self):
        return self.Authorization


@dataclass
class AuthenticationRequest:
    """Authentication request model"""

    user: str
    authcode: str


@dataclass
class User:
    """User model."""

    principal: str


@dataclass
class LoginResponse:
    """Login response model"""

    access_token: str

    def as_dict(self):
        return {"access_token": self.access_token}



from aws_packages.auth.auth_backend_base import AuthBackendBase

AUTH_SETTINGS = {"AUTH_BACKEND": None}


def get_auth_backend():
    return AUTH_SETTINGS["AUTH_BACKEND"]


def set_auth_backend(auth_backend: AuthBackendBase):
    AUTH_SETTINGS["AUTH_BACKEND"] = auth_backend



from aws_packages.auth.models import AuthenticationRequest, User


class AuthBackendBase:
    """Base class for authentication backends"""

    def authenticate(self, request: AuthenticationRequest) -> User:
        """Authenticate the request and returns an authenticated User"""

        raise NotImplementedError("authenticate not implemented")

    def authenticate_with_token(self, token: str) -> bool:
        """Authenticate user using jwt token and return status"""
        raise NotImplementedError("authenticate_with_token not implemented")

    def get_access_token(self, user: User) -> str:
        """Get the access for the given user"""
        raise NotImplementedError("get_access_token not implemented")

    def get_refresh_token(self, user: User, token: str) -> str:
        """Get the refresh token for the given user"""
        raise NotImplementedError("get_refresh_token not implemented")



"""Auth middleware."""

from aws_lambda_powertools.event_handler import APIGatewayRestResolver, Response
from aws_lambda_powertools.event_handler.exceptions import UnauthorizedError
from aws_lambda_powertools.event_handler.middlewares import NextMiddleware

from aws_packages.auth.auth_backend_icp import icp_auth_backend
from aws_packages.auth.models import AuthorizationRequest

from .exceptions import JWTAuthError


def icp_login_required(
    app: APIGatewayRestResolver, next_middleware: NextMiddleware
) -> Response:
    """Login required middleware

    :param app: application instance
    :param next_middleware: next middleware
    :return: Response object after the middleware has been applied
    """

    request = AuthorizationRequest(**app.current_event.headers)
    try:
        icp_auth_backend.authenticate_with_token(request.token)
    except JWTAuthError:
        raise UnauthorizedError(f"Unauthorized")
    return next_middleware(app)


def solana_login_required(
    app: APIGatewayRestResolver, next_middleware: NextMiddleware
) -> Response:
    """Login required middleware

    :param app: application instance
    :param next_middleware: next middleware
    :return: Response object after the middleware has been applied
    """

    request = AuthorizationRequest(**app.current_event.headers)
    try:
        icp_auth_backend.authenticate_with_token(request.token)
    except JWTAuthError:
        raise UnauthorizedError(f"Unauthorized")
    return next_middleware(app)



"""Authentication backends.

"""

import os

from aws_packages.auth.auth_backend_base import AuthBackendBase
from aws_packages.auth.models import AuthenticationRequest, User
from aws_packages.auth.tokens import AccessToken


class SolanaAuthBackend(AuthBackendBase):
    """Authentication backend with Solana Canister"""

    def __init__(self, url: str):
        self._url = url

    def authenticate(self, login_request: AuthenticationRequest) -> User:
        """Authenticate the request and returns an authenticated User"""

        # solana sdk call and authenticate or raise exception

        return User(principal=login_request.user)

    def authenticate_with_token(self, token: str):
        """Authenticate user using jwt token and return status"""
        print("token : ", token)
        AccessToken(token_string=token)

    def get_access_token(self, user: User) -> str:
        """Get the access for the given user"""
        token = AccessToken()
        return token.get_access_token(user)


solana_auth_backend = SolanaAuthBackend(
    url="https://ic0.app",
)

if __name__ == "__main__":
    icp_auth_backend = SolanaAuthBackend(
        url="https://ic0.app",
    )

    request = AuthenticationRequest(
        user="4cay5-ew3bs-vr6yl-7iffu-67doc-l655v-dluy7-qplpx-7pkio-er5rt-uqe",
        authcode="",
    )
    user = solana_auth_backend.authenticate(request)
    print(user)



from calendar import timegm
from datetime import datetime, timezone


def datetime_to_epoch(dt: datetime) -> int:
    return timegm(dt.utctimetuple())


def aware_utcnow(use_tz=False) -> datetime:
    dt = datetime.now(tz=timezone.utc)
    if not use_tz:
        dt = dt.replace(tzinfo=None)

    return dt


def get_utc_now() -> int:
    return datetime_to_epoch(aware_utcnow())



import os
from datetime import timedelta

import jwt
from models import User

from aws_packages.time import aware_utcnow, datetime_to_epoch

from .exceptions import JWTAuthError

DEFAULT_SECRET_KEY = "27d621e9bc55e6c659842904982abf06d89123c844e4d8bc62060ccd6536c360"
SECRET_KEY = os.environ.get("SECRET_KEY", DEFAULT_SECRET_KEY)


class Token:
    secret_key = SECRET_KEY
    lifetime = timedelta(minutes=5)

    def __init__(self, token_string=None):
        self._token_string = token_string
        if self._token_string:
            self.validate_jwt_token()
        self._exp = aware_utcnow() + self.lifetime
        self._jwt_payload = {"exp": self.get_exp()}

    def get_exp(self):
        return datetime_to_epoch(self._exp)

    def set_exp(self, value):
        self._jwt_payload["exp"] = value

    def get_access_token(self, user: User):
        self._jwt_payload["user_id"] = user.principal

        encoded = jwt.encode(
            self._jwt_payload,
            self.secret_key,
            algorithm="HS256",
        )
        return encoded

    def validate_jwt_token(self):
        # check the token validity
        decoded_jwt = jwt.decode(
            self._token_string, self.secret_key, algorithms=["HS256"]
        )
        print(decoded_jwt)

        jwt_expiration = decoded_jwt["exp"]
        timestamp_now = datetime_to_epoch(aware_utcnow())
        lifetime_remaining = jwt_expiration - timestamp_now
        print(lifetime_remaining)
        if lifetime_remaining < 0:
            raise JWTAuthError("JWT Token Expired")


class AccessToken(Token):
    lifetime = timedelta(days=30)


if __name__ == "__main__":
    user = User(principal="test_user")

    token = AccessToken(token_string=None)
    new_jwt_token = token.get_access_token(user)

    print(new_jwt_token)
